@@ -15,44 +15,35 @@
 #
 
 import argparse
+import json
 import logging
 import logging.handlers
 import sys
 import time
 
-from pyquarantine import QuarantineMilter, setup_milter
+from pyquarantine.config import get_milter_config, ActionConfig
+from pyquarantine.storage import Quarantine
 from pyquarantine import __version__ as version
 
 
-def _get_quarantine(quarantines, name):
+def _get_quarantine(quarantines, name, debug):
     try:
-        quarantine = next((q for q in quarantines if q.name == name))
+        quarantine = next((q for q in quarantines if q["name"] == name))
     except StopIteration:
-        raise RuntimeError("invalid quarantine 'name'")
-    return quarantine
+        raise RuntimeError(f"invalid quarantine '{name}'")
+    return Quarantine(ActionConfig(quarantine), [], debug)
 
 
-def _get_storage(quarantines, name):
-    quarantine = _get_quarantine(quarantines, name)
-    storage = quarantine.get_storage()
-    if not storage:
-        raise RuntimeError(
-                "storage type is set to NONE")
-    return storage
-
-
-def _get_notification(quarantines, name):
-    quarantine = _get_quarantine(quarantines, name)
-    notification = quarantine.get_notification()
+def _get_notification(quarantines, name, debug):
+    notification = _get_quarantine(quarantines, name, debug).notification
     if not notification:
         raise RuntimeError(
                 "notification type is set to NONE")
     return notification
 
 
-def _get_whitelist(quarantines, name):
-    quarantine = _get_quarantine(quarantines, name)
-    whitelist = quarantine.get_whitelist()
+def _get_whitelist(quarantines, name, debug):
+    whitelist = _get_quarantine(quarantines, name, debug).whitelist
     if not whitelist:
         raise RuntimeError(
                 "whitelist type is set to NONE")
@@ -102,34 +93,35 @@ def print_table(columns, rows):
 
 def list_quarantines(quarantines, args):
     if args.batch:
-        print("\n".join([q.name for q in quarantines]))
+        print("\n".join([q["name"] for q in quarantines]))
     else:
         qlist = []
         for q in quarantines:
-            storage = q.get_storage()
-            if storage:
-                storage_type = q.get_storage().storage_type
-            else:
-                storage_type = "NONE"
+            cfg = q["options"]
+            storage_type = cfg["store"]["type"]
 
-            notification = q.get_notification()
-            if notification:
-                notification_type = q.get_notification().notification_type
+            if "notify" in cfg:
+                notification_type = cfg["notify"]["type"]
             else:
                 notification_type = "NONE"
 
-            whitelist = q.get_whitelist()
-            if whitelist:
-                whitelist_type = q.get_whitelist().whitelist_type
+            if "whitelist" in cfg:
+                whitelist_type = cfg["whitelist"]["type"]
             else:
                 whitelist_type = "NONE"
 
+            if "milter_action" in cfg:
+                milter_action = cfg["milter_action"]
+            else:
+                milter_action = "NONE"
+
             qlist.append({
-                "name": q.name,
+                "name": q["name"],
                 "storage": storage_type,
                 "notification": notification_type,
                 "whitelist": whitelist_type,
-                "action": q.action})
+                "action": milter_action})
+
         print_table(
             [("Name", "name"),
              ("Storage", "storage"),
@@ -141,8 +133,8 @@ def list_quarantines(quarantines, args):
 
 
 def list_quarantine_emails(quarantines, args):
-    logger = logging.getLogger(__name__)
-    storage = _get_storage(quarantines, args.quarantine)
+    storage = _get_quarantine(quarantines, args.quarantine, args.debug).storage
+
     # find emails and transform some metadata values to strings
     rows = []
     emails = storage.find(
@@ -150,21 +142,21 @@ def list_quarantine_emails(quarantines, args):
     for storage_id, metadata in emails.items():
         row = emails[storage_id]
         row["storage_id"] = storage_id
-        row["date"] = time.strftime(
+        row["timestamp"] = time.strftime(
             '%Y-%m-%d %H:%M:%S',
             time.localtime(
-                metadata["date"]))
+                metadata["timestamp"]))
         row["mailfrom"] = metadata["mailfrom"]
         row["recipient"] = metadata["recipients"].pop(0)
-        if "subject" not in emails[storage_id]["headers"].keys():
-            emails[storage_id]["headers"]["subject"] = ""
-        row["subject"] = emails[storage_id]["headers"]["subject"][:60].strip()
+        if "subject" not in emails[storage_id]:
+            emails[storage_id]["subject"] = ""
+        row["subject"] = emails[storage_id]["subject"][:60].strip()
         rows.append(row)
 
         if metadata["recipients"]:
             row = {
                 "storage_id": "",
-                "date": "",
+                "timestamp": "",
                 "mailfrom": "",
                 "recipient": metadata["recipients"].pop(0),
                 "subject": ""
@@ -177,9 +169,11 @@ def list_quarantine_emails(quarantines, args):
         return
 
     if not emails:
-        logger.info(f"quarantine '{args.quarantine}' is empty")
+        print(f"quarantine '{args.quarantine}' is empty")
+        return
+
     print_table(
-        [("Quarantine-ID", "storage_id"), ("Date", "date"),
+        [("Quarantine-ID", "storage_id"), ("When", "timestamp"),
          ("From", "mailfrom"), ("Recipient(s)", "recipient"),
          ("Subject", "subject")],
         rows
@@ -187,8 +181,7 @@ def list_quarantine_emails(quarantines, args):
 
 
 def list_whitelist(quarantines, args):
-    logger = logging.getLogger(__name__)
-    whitelist = _get_whitelist(quarantines, args.quarantine)
+    whitelist = _get_whitelist(quarantines, args.quarantine, args.debug)
 
     # find whitelist entries
     entries = whitelist.find(
@@ -196,16 +189,15 @@ def list_whitelist(quarantines, args):
         recipients=args.recipients,
         older_than=args.older_than)
     if not entries:
-        logger.info(
-            f"whitelist of quarantine '{args.quarantine}' is empty")
+        print(f"whitelist of quarantine '{args.quarantine}' is empty")
         return
 
     # transform some values to strings
-    for entry_id, entry in entries.items():
-        entries[entry_id]["permanent_str"] = str(entry["permanent"])
-        entries[entry_id]["created_str"] = entry["created"].strftime(
+    for eid, entry in entries.items():
+        entries[eid]["permanent_str"] = str(entry["permanent"])
+        entries[eid]["created_str"] = entry["created"].strftime(
             '%Y-%m-%d %H:%M:%S')
-        entries[entry_id]["last_used_str"] = entry["last_used"].strftime(
+        entries[eid]["last_used_str"] = entry["last_used"].strftime(
             '%Y-%m-%d %H:%M:%S')
 
     print_table(
@@ -220,10 +212,10 @@ def list_whitelist(quarantines, args):
 
 def add_whitelist_entry(quarantines, args):
     logger = logging.getLogger(__name__)
-    whitelist = _get_whitelist(quarantines, args.quarantine)
+    whitelist = _get_whitelist(quarantines, args.quarantine, args.debug)
 
     # check existing entries
-    entries = whitelist.check(args.mailfrom, args.recipient)
+    entries = whitelist.check(args.mailfrom, args.recipient, logger)
     if entries:
         # check if the exact entry exists already
         for entry in entries.values():
@@ -234,11 +226,11 @@ def add_whitelist_entry(quarantines, args):
 
         if not args.force:
             # the entry is already covered by others
-            for entry_id, entry in entries.items():
-                entries[entry_id]["permanent_str"] = str(entry["permanent"])
-                entries[entry_id]["created_str"] = entry["created"].strftime(
+            for eid, entry in entries.items():
+                entries[eid]["permanent_str"] = str(entry["permanent"])
+                entries[eid]["created_str"] = entry["created"].strftime(
                     '%Y-%m-%d %H:%M:%S')
-                entries[entry_id]["last_used_str"] = entry["last_used"].strftime(
+                entries[eid]["last_used_str"] = entry["last_used"].strftime(
                     '%Y-%m-%d %H:%M:%S')
             print_table(
                 [
@@ -255,42 +247,47 @@ def add_whitelist_entry(quarantines, args):
 
     # add entry to whitelist
     whitelist.add(args.mailfrom, args.recipient, args.comment, args.permanent)
-    logger.info("whitelist entry added successfully")
+    print("whitelist entry added successfully")
 
 
 def delete_whitelist_entry(quarantines, args):
-    logger = logging.getLogger(__name__)
-    whitelist = _get_whitelist(quarantines, args.quarantine)
+    whitelist = _get_whitelist(quarantines, args.quarantine, args.debug)
     whitelist.delete(args.whitelist_id)
-    logger.info("whitelist entry deleted successfully")
+    print("whitelist entry deleted successfully")
 
 
 def notify(quarantines, args):
-    logger = logging.getLogger(__name__)
-    quarantine = _get_quarantine(quarantines, args.quarantine)
+    quarantine = _get_quarantine(quarantines, args.quarantine, args.debug)
     quarantine.notify(args.quarantine_id, args.recipient)
-    logger.info("notification sent successfully")
+    print("notification sent successfully")
 
 
 def release(quarantines, args):
     logger = logging.getLogger(__name__)
-    quarantine = _get_quarantine(quarantines, args.quarantine)
-    quarantine.release(args.quarantine_id, args.recipient)
-    logger.info("quarantined email released successfully")
+    quarantine = _get_quarantine(quarantines, args.quarantine, args.debug)
+    rcpts = quarantine.release(args.quarantine_id, args.recipient)
+    rcpts = ", ".join(rcpts)
+    logger.info(
+        f"{args.quarantine}: released message with id {args.quarantine_id} "
+        f"for {rcpts}")
 
 
 def delete(quarantines, args):
-    logger = logging.getLogger(__name__)
-    storage = _get_storage(quarantines, args.quarantine)
+    storage = _get_quarantine(quarantines, args.quarantine, args.debug).storage
     storage.delete(args.quarantine_id, args.recipient)
-    logger.info("quarantined email deleted successfully")
+    print("quarantined message deleted successfully")
 
 
 def get(quarantines, args):
-    storage = _get_storage(quarantines, args.quarantine)
-    fp, _ = storage.get_mail(args.quarantine_id)
-    print(fp.read().decode())
-    fp.close()
+    storage = _get_quarantine(quarantines, args.quarantine, args.debug).storage
+    data = storage.get_mail_bytes(args.quarantine_id)
+    sys.stdout.buffer.write(data)
+
+
+def metadata(quarantines, args):
+    storage = _get_quarantine(quarantines, args.quarantine, args.debug).storage
+    metadata = storage.get_metadata(args.quarantine_id)
+    print(json.dumps(metadata))
 
 
 class StdErrFilter(logging.Filter):
@@ -304,6 +301,9 @@ class StdOutFilter(logging.Filter):
 
 
 def main():
+    python_version = ".".join([str(v) for v in sys.version_info[0:3]])
+    python_version = f"{python_version}-{sys.version_info[3]}"
+
     "PyQuarantine command-line interface."
     # parse command line
     def formatter_class(prog): return argparse.HelpFormatter(
@@ -312,10 +312,8 @@ def main():
         description="PyQuarantine CLI",
         formatter_class=formatter_class)
     parser.add_argument(
-        "-c", "--config",
-        help="Config files to read.",
-        nargs="+", metavar="CFG",
-        default=QuarantineMilter.get_cfg_files())
+        "-c", "--config", help="Config file to read.",
+        default="/etc/pyquarantine/pyquarantine.conf")
     parser.add_argument(
         "-d", "--debug",
         help="Log debugging messages.",
@@ -324,7 +322,7 @@ def main():
         "-v", "--version",
         help="Print version.",
         action="version",
-        version=f"%(prog)s ({version})")
+        version=f"%(prog)s {version} (python {python_version})")
     parser.set_defaults(syslog=False)
     subparsers = parser.add_subparsers(
         dest="command",
@@ -343,133 +341,145 @@ def main():
     list_parser.set_defaults(func=list_quarantines)
 
     # quarantine command group
-    quarantine_parser = subparsers.add_parser(
+    quar_parser = subparsers.add_parser(
         "quarantine",
         description="Manage quarantines.",
         help="Manage quarantines.",
         formatter_class=formatter_class)
-    quarantine_parser.add_argument(
+    quar_parser.add_argument(
         "quarantine",
         metavar="QUARANTINE",
         help="Quarantine name.")
-    quarantine_subparsers = quarantine_parser.add_subparsers(
+    quar_subparsers = quar_parser.add_subparsers(
         dest="command",
         title="Quarantine commands")
-    quarantine_subparsers.required = True
+    quar_subparsers.required = True
     # quarantine list command
-    quarantine_list_parser = quarantine_subparsers.add_parser(
+    quar_list_parser = quar_subparsers.add_parser(
         "list",
         description="List emails in quarantines.",
         help="List emails in quarantine.",
         formatter_class=formatter_class)
-    quarantine_list_parser.add_argument(
+    quar_list_parser.add_argument(
         "-f", "--from",
         dest="mailfrom",
         help="Filter emails by from address.",
         default=None,
         nargs="+")
-    quarantine_list_parser.add_argument(
+    quar_list_parser.add_argument(
         "-t", "--to",
         dest="recipients",
         help="Filter emails by recipient address.",
         default=None,
         nargs="+")
-    quarantine_list_parser.add_argument(
+    quar_list_parser.add_argument(
         "-o", "--older-than",
         dest="older_than",
         help="Filter emails by age (days).",
         default=None,
         type=float)
-    quarantine_list_parser.add_argument(
+    quar_list_parser.add_argument(
         "-b", "--batch",
-        help="Print results using only email quarantine IDs, each on a new line.",
+        help="Print results using only email quarantine IDs, "
+             "each on a new line.",
         action="store_true")
-    quarantine_list_parser.set_defaults(func=list_quarantine_emails)
+    quar_list_parser.set_defaults(func=list_quarantine_emails)
     # quarantine notify command
-    quarantine_notify_parser = quarantine_subparsers.add_parser(
+    quar_notify_parser = quar_subparsers.add_parser(
         "notify",
         description="Notify recipient about email in quarantine.",
         help="Notify recipient about email in quarantine.",
         formatter_class=formatter_class)
-    quarantine_notify_parser.add_argument(
+    quar_notify_parser.add_argument(
         "quarantine_id",
         metavar="ID",
         help="Quarantine ID.")
-    quarantine_notify_parser_group = quarantine_notify_parser.add_mutually_exclusive_group(
+    quar_notify_parser_grp = quar_notify_parser.add_mutually_exclusive_group(
         required=True)
-    quarantine_notify_parser_group.add_argument(
+    quar_notify_parser_grp.add_argument(
         "-t", "--to",
         dest="recipient",
         help="Release email for one recipient address.")
-    quarantine_notify_parser_group.add_argument(
+    quar_notify_parser_grp.add_argument(
         "-a", "--all",
         help="Release email for all recipients.",
         action="store_true")
-    quarantine_notify_parser.set_defaults(func=notify)
+    quar_notify_parser.set_defaults(func=notify)
     # quarantine release command
-    quarantine_release_parser = quarantine_subparsers.add_parser(
+    quar_release_parser = quar_subparsers.add_parser(
         "release",
         description="Release email from quarantine.",
         help="Release email from quarantine.",
         formatter_class=formatter_class)
-    quarantine_release_parser.add_argument(
+    quar_release_parser.add_argument(
         "quarantine_id",
         metavar="ID",
         help="Quarantine ID.")
-    quarantine_release_parser.add_argument(
+    quar_release_parser.add_argument(
         "-n",
         "--disable-syslog",
         dest="syslog",
         help="Disable syslog messages.",
         action="store_false")
-    quarantine_release_parser_group = quarantine_release_parser.add_mutually_exclusive_group(
+    quar_release_parser_grp = quar_release_parser.add_mutually_exclusive_group(
         required=True)
-    quarantine_release_parser_group.add_argument(
+    quar_release_parser_grp.add_argument(
         "-t", "--to",
         dest="recipient",
         help="Release email for one recipient address.")
-    quarantine_release_parser_group.add_argument(
+    quar_release_parser_grp.add_argument(
         "-a", "--all",
         help="Release email for all recipients.",
         action="store_true")
-    quarantine_release_parser.set_defaults(func=release)
+    quar_release_parser.set_defaults(func=release)
     # quarantine delete command
-    quarantine_delete_parser = quarantine_subparsers.add_parser(
+    quar_delete_parser = quar_subparsers.add_parser(
         "delete",
         description="Delete email from quarantine.",
         help="Delete email from quarantine.",
         formatter_class=formatter_class)
-    quarantine_delete_parser.add_argument(
+    quar_delete_parser.add_argument(
         "quarantine_id",
         metavar="ID",
         help="Quarantine ID.")
-    quarantine_delete_parser.add_argument(
+    quar_delete_parser.add_argument(
         "-n", "--disable-syslog",
         dest="syslog",
         help="Disable syslog messages.",
         action="store_false")
-    quarantine_delete_parser_group = quarantine_delete_parser.add_mutually_exclusive_group(
+    quar_delete_parser_grp = quar_delete_parser.add_mutually_exclusive_group(
         required=True)
-    quarantine_delete_parser_group.add_argument(
+    quar_delete_parser_grp.add_argument(
         "-t", "--to",
         dest="recipient",
         help="Delete email for one recipient address.")
-    quarantine_delete_parser_group.add_argument(
+    quar_delete_parser_grp.add_argument(
         "-a", "--all",
         help="Delete email for all recipients.",
         action="store_true")
-    quarantine_delete_parser.set_defaults(func=delete)
+    quar_delete_parser.set_defaults(func=delete)
     # quarantine get command
-    quarantine_get_parser = quarantine_subparsers.add_parser(
+    quar_get_parser = quar_subparsers.add_parser(
         "get",
         description="Get email from quarantine.",
         help="Get email from quarantine",
         formatter_class=formatter_class)
-    quarantine_get_parser.add_argument(
+    quar_get_parser.add_argument(
         "quarantine_id",
         metavar="ID",
         help="Quarantine ID.")
-    quarantine_get_parser.set_defaults(func=get)
+    quar_get_parser.set_defaults(func=get)
+    # quarantine metadata command
+    quar_metadata_parser = quar_subparsers.add_parser(
+        "metadata",
+        description="Get metadata of email from quarantine.",
+        help="Get metadata of email from quarantine",
+        formatter_class=formatter_class)
+    quar_metadata_parser.add_argument(
+        "quarantine_id",
+        metavar="ID",
+        help="Quarantine ID.")
+    quar_metadata_parser.set_defaults(func=metadata)
 
     # whitelist command group
     whitelist_parser = subparsers.add_parser(
@@ -536,7 +546,8 @@ def main():
         action="store_true")
     whitelist_add_parser.add_argument(
         "--force",
-        help="Force adding an entry, even if already covered by another entry.",
+        help="Force adding an entry, "
+             "even if already covered by another entry.",
         action="store_true")
     whitelist_add_parser.set_defaults(func=add_whitelist_entry)
     # whitelist delete command
@@ -578,14 +589,26 @@ def main():
     root_logger.addHandler(stderrhandler)
     logger = logging.getLogger(__name__)
 
-    # try to generate milter configs
     try:
-        setup_milter(
-            cfg_files=args.config, test=True)
-    except RuntimeError as e:
-        logger.error(e)
+        logger.debug("read milter configuration")
+        cfg = get_milter_config(args.config, raw=True)
+        if "rules" not in cfg or not cfg["rules"]:
+            raise RuntimeError("no rules configured")
+
+        for rule in cfg["rules"]:
+            if "actions" not in rule or not rule["actions"]:
+                raise RuntimeError(
+                    f"{rule['name']}: no actions configured")
+    except (RuntimeError, AssertionError) as e:
+        logger.error(f"config error: {e}")
         sys.exit(255)
 
+    quarantines = []
+    for rule in cfg["rules"]:
+        for action in rule["actions"]:
+            if action["type"] == "quarantine":
+                quarantines.append(action)
+
     if args.syslog:
         # setup syslog
         sysloghandler = logging.handlers.SysLogHandler(
@@ -602,7 +625,7 @@ def main():
 
     # call the commands function
     try:
-        args.func(QuarantineMilter.quarantines, args)
+        args.func(quarantines, args)
     except RuntimeError as e:
         logger.error(e)
         sys.exit(1)
